/* zebra_mroute.h
 * Copyright (C) 2016 Cumulus Networks, Inc.
 * Donald Sharp
 *
 * This file is part of Quagga.
 *
 * Quagga is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * Quagga is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; see the file COPYING; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */

#ifndef __ZEBRA_MROUTE_H__
#define __ZEBRA_MROUTE_H__

#include "zebra/zserv.h"

#ifdef __cplusplus
extern "C" {
#endif

struct mcast_route_data {
	struct prefix_sg sg;
	unsigned int ifindex;
	unsigned long long lastused;
};

void zebra_ipmr_route_stats(ZAPI_HANDLER_ARGS);

#ifdef __cplusplus
}
#endif

#endif
