/*
 * This file is part of Quagga.
 *
 * Quagga is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * Quagga is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; see the file COPYING; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */

#ifndef _QUAGGA_ASSERT_H
#define _QUAGGA_ASSERT_H

#ifdef __cplusplus
extern "C" {
#endif

extern void _zlog_assert_failed(const char *assertion, const char *file,
				unsigned int line, const char *function)
	__attribute__((noreturn));

#if defined(__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
#define __ASSERT_FUNCTION    __func__
#elif defined(__GNUC__)
#define __ASSERT_FUNCTION    __FUNCTION__
#else
#define __ASSERT_FUNCTION    NULL
#endif

#define zassert(EX)                                                            \
	((void)((EX) ? 0 : (_zlog_assert_failed(#EX, __FILE__, __LINE__,       \
						__ASSERT_FUNCTION),            \
			    0)))

#undef assert
#define assert(EX) zassert(EX)

#ifdef __cplusplus
}
#endif

#endif /* _QUAGGA_ASSERT_H */
